@@ -1,4 +1,6 @@
 """Main module."""
+from transformers import logging
+logging.set_verbosity_error()
 import re
 import string
 import ast
@@ -6,25 +8,34 @@ import importlib_resources
 from collections import Counter
 from transformers import pipeline
 from symspellpy import SymSpell, Verbosity
+from metaphone import doublemetaphone
 import pkg_resources
 
 
+### Load in project resources
+# Full word list
 ocrfixr = importlib_resources.files("ocrfixr")
 word_set = (ocrfixr / "data" / "SCOWL_70.txt").read_text().split()
 word_set = set(word_set)
 
-common_scannos = (ocrfixr / "data" / "common_scannos.txt").read_text()
+# List of words NOT in SCOWL 70 that we should ignore anyways
+ignore_misreads = (ocrfixr / "data" / "ignore_these_misspells.txt").read_text().split()
+ignore_set_from_pkg = set(ignore_misreads)
+
+# dict of common scannos to check for (bypasses the context check, since these are clear mappings)
+common_scannos = (ocrfixr / "data" / "scannos_common.txt").read_text()
 common_scannos = ast.literal_eval(common_scannos)
 common = set(common_scannos)
 
-
-stealth_scannos = (ocrfixr / "data" / "stealth_scannos.txt").read_text()
+# dict of specifically tricky scannos to check for (bypasses the context check, since these are clear mappings)
+stealth_scannos = (ocrfixr / "data" / "scannos_stealth.txt").read_text()
 stealth_scannos = ast.literal_eval(stealth_scannos)
 stealth = set(stealth_scannos)
 
-ignore_misreads = (ocrfixr / "data" / "ignore_these.txt").read_text().split()
-ignore_set_from_pkg = set(ignore_misreads)
-
+# dict of OCRfixr suggestions that are known to be bad. This list prevents them from ever being suggested.
+ignore_suggestions = (ocrfixr / "data" / "ignore_these_suggestions.txt").read_text()
+ignore_suggestions = ast.literal_eval(ignore_suggestions)
+#ignore_suggestions = set(ignore_suggestions)
 
 # setup symspell spellchecker parameters
 sym_spell = SymSpell(max_dictionary_edit_distance=2, prefix_length=7)
@@ -40,7 +51,7 @@ unmasker = pipeline('fill-mask', model='bert-base-uncased', top_k=30)
 
 
 class spellcheck:                       
-    def __init__(self, text, changes_by_paragraph = "F", return_fixes = "F", ignore_words = None, interactive = "F", common_scannos = "T", top_k = 15):
+    def __init__(self, text, changes_by_paragraph = "F", return_fixes = "F", ignore_words = None, interactive = "F", common_scannos = "T", top_k = 15, return_context = "F"):
         self.text = text
         self.changes_by_paragraph = changes_by_paragraph
         self.return_fixes = return_fixes
@@ -48,6 +59,7 @@ class spellcheck:
         self.interactive = interactive
         self.common_scannos = common_scannos
         self.top_k = top_k
+        self.return_context = return_context
 
         
 ### DEFINE ALL HELPER FUNCTIONS
@@ -69,7 +81,7 @@ class spellcheck:
         # Also, drop all words with trailing numbers flanked by punctuation, indicating a footnote reference (money.4, item[1]) rather than a misspelling
         # Also, drop any 1-character "words" 
 
-        no_hyphens = re.compile(".*-.*|.*'.*|[0-9]+")
+        no_hyphens = re.compile(".*-.*|.*'.*|.*’.*|[0-9]+")
         no_caps = re.compile('[^A-Z][a-z0-9]{1,}')
         no_footnotes = re.compile('.*[0-9]{1,}[^A-z]?$')
         no_list_items = re.compile('.*\\)|.*\\]')
@@ -78,7 +90,7 @@ class spellcheck:
         words = [x for x in tokens if not no_hyphens.match(x) and no_caps.match(x) and not no_footnotes.match(x) and not no_list_items.match(x)]
 
         # then, remove punct from each remaining token (such as trailing commas, periods, quotations ('' & ""), but KEEPING contractions). 
-        no_punctuation = [l.strip(string.punctuation) for l in words]
+        no_punctuation = [l.strip(string.punctuation+"”“’‘") for l in words]
         words_to_check = [x for x in no_punctuation if len(x) > 1 and not all_nums.match(x)]
         
         
@@ -121,10 +133,23 @@ class spellcheck:
     # Return the list of possible spell-check options. These will be used to look for matches against BERT context suggestions
     def __SUGGEST_SPELLCHECK(self, text):
         suggested_words = []
-        for i in sym_spell.lookup(text, Verbosity.CLOSEST, max_edit_distance=2):
-            term = getattr(i, "term")
-            if len(term) > 1:
-                suggested_words.append(term)
+        
+        # Confirm word isn't a mashup ("anhour" --> "an hour")
+        Num_spaces = []
+        for i in sym_spell.lookup_compound(text, max_edit_distance=0):
+            Num_spaces.append(getattr(i, "term"))
+
+        # If it is not flagged as a possible mashup, then ask for plausible replacement words for the full misspelled word                
+        if str(Num_spaces).count(' ') == 0:
+            for i in sym_spell.lookup(text, Verbosity.CLOSEST, max_edit_distance=2):
+                term = getattr(i, "term")
+                if len(term) > 1:
+                    suggested_words.append(term)
+        
+        # If symspell suggests that the misspell should actually be be 2 words, then pass the multi-word phrase
+        else:
+            suggested_words.append(Num_spaces.pop())
+            
         return(suggested_words)
         
     
@@ -216,9 +241,9 @@ class spellcheck:
         content.rowconfigure(1, weight=1)
         
         root.mainloop()
+
         
-        
-    # Creates a dict of valid replacements for misspellings. If bert and pyspellcheck do not have a match for a given misspelling, it makes no changes to the word.
+    # Creates a dict of valid replacements for misspellings. If bert and symspell do not have a match for a given misspelling, it makes no changes to the word.
     # When common_scannos is activated, that limited list of words bypass the spellcheck/context check
     def _FIND_REPLACEMENTS(self, misreads):
         SC = [] 
@@ -229,7 +254,7 @@ class spellcheck:
         for i in misreads:
             # if misread is a common scanno, then add that entry to a separate dict that will be merged back in later. This bypasses the BERT check step.
             if self.common_scannos == "T" and i in common:
-                add_scanno = dict((k, v) for k, v in common_scannos.items() if k in i)
+                add_scanno = dict((k, v) for k, v in common_scannos.items() if k == i)
                 common_scanno_fixes.update(add_scanno)
             
             # for stealth scannos - these are valid (yet incorrect) words. So, instead of SUGGEST_SPELLCHECK (which would return the same word supplied), take the value from the stealth_scanno dict, which is the desired word to check for in BERT context (arid --> and)
@@ -238,18 +263,34 @@ class spellcheck:
                 bert.append(self.__SUGGEST_BERT(text = self.__SET_MASK(i,'[MASK]', self.text), 
                                                 number_to_return = self.top_k))
             
-            # for all other unrecognized words, get all spellcheck suggestions from textblob
+            # for all other unrecognized words, get all spellcheck suggestions from symspell
             else:
                 spellcheck = self.__SUGGEST_SPELLCHECK(i)
-                # Make sure there is a valid spellcheck suggestion (textBlob returns the original string if not)
+                # Make sure there is a valid spellcheck suggestion (symspell returns the original string if not)
                 if spellcheck == i:
-                    # if none, remove from misreads (ie. dont bother checking BERT context - it won't get used)
+                    # if no spellcheck suggestion given, remove from misreads (ie. dont bother checking BERT context - it won't get used)
                     misreads.delete(i)
-                else:
-                    # otherwise, do the spellcheck + context check
-                    SC.append(spellcheck)
-                    bert.append(self.__SUGGEST_BERT(text = self.__SET_MASK(i,'[MASK]', self.text), 
-                                                    number_to_return = self.top_k))
+                else: 
+                    # tee up symspell suggestion for comparison to BERT suggestion
+                    SC.append(spellcheck)  
+                    
+                    # For multi-word phrases, the BERT context check will feed in the first word into the text, then confirm whether second word fits the context of the sentence. If so, the two word phrase will be accepted as a valid correction
+                    if len(spellcheck) == 1 and str(spellcheck).count(' ') == 1:
+                        mw = ''.join(spellcheck)
+                        fw = re.findall("^[^\s]+", mw).pop()
+                        SB = self.__SUGGEST_BERT(text = self.__SET_MASK(i, fw + ' [MASK]', self.text), 
+                                                        number_to_return = self.top_k)     
+                        # Tack the first word onto the results for each BERT context suggestion. These are compared against the multi-word phrase provided by sympell
+                        SBi = []
+                        for i in SB:
+                            SBi.append(fw + ' ' + i )
+                        
+                        bert.append(SBi)
+
+                    else:    
+                        # otherwise, just mask the misspelled word for BERT context check, which will be compared against symspell
+                        bert.append(self.__SUGGEST_BERT(text = self.__SET_MASK(i,'[MASK]', self.text), 
+                                                        number_to_return = self.top_k))
     
         # then, see if spellcheck & bert overlap
         # if they do, set that value for the find-replace dict
@@ -262,23 +303,50 @@ class spellcheck:
         while x < len(bert):
             overlap = set(bert[x]) & set(SC[x])
             corr.append(overlap)
-            # if there is a single word that is both in context and pyspellcheck - update with that word
+            # if there is a single word that is both in context and symspellpy - update with that word
             if len(overlap) == 1:
                 corr[x] = self.__LIST_TO_STR(corr[x])                
             # if no overlapping candidates OR > 1 candidate, keep misread as is
             else:
                 corr[x] = ""
             x = x+1
-    
+            
+            
         fixes = dict(zip(misreads, corr))
+        
+        try:
+            for key, value in fixes.copy().items():
+                # if it's a simple "remove an 's' from the end", (kissings --> kissing) then delete that fix
+                if value + "s" == key:
+                    del fixes[key]
+                # if it's an o -> e ending fix (bo --> be), ignore the soundex check
+                elif key[len(key)-1] == "o" and value[len(value)-1] == "e":
+                    pass
+                # ignore soundex check for mashups (anhour --> an hour). Found by looking for a space (' ') in the replacement suggestion                
+                elif value.count(' ') == 1:
+                    pass
+                # Check whether the find-replace candidate is a homophone - these suggestions are ignored, to avoid flagging intentional (stylistic) homophones (ie. without / widout)
+                # soundex check = double metaphone
+                elif doublemetaphone(key)[0] == doublemetaphone(value)[0]:
+                    del fixes[key]
+        # If soundex can't parse the character, just skip the check
+        except Exception:
+            pass
+          
+        # Add the common scannos to the mix
         fixes.update(common_scanno_fixes)
     
     
         # Remove all dict entries where replacement is same as initial problematic text
-        keys = [k for k, v in fixes.items() if k == v]
-        for x in keys:
+        no_change = [k for k, v in fixes.items() if k == v]
+        for x in no_change:
             del fixes[x]
                 
+        # Remove all dict entries that are in the list of known bad suggestions
+        overlap = dict(fixes.items() & ignore_suggestions.items())
+        for x in overlap:
+            del fixes[x]
+        
             
         if self.interactive == "T":
             # Remove all dict entries with "" values (ie. no suggested change) --- this cleans up the suggestions the user sees
@@ -324,8 +392,19 @@ class spellcheck:
             fixes = self._FIND_REPLACEMENTS(misreads)
             correction = self._MULTI_REPLACE(fixes)
             # for any text that has no updates, remove from changes_by_paragraph output
-            if self.changes_by_paragraph == "T" and len(fixes) == 0:
-                full_results = []
+            if self.changes_by_paragraph == "T":
+                if len(fixes) == 0:
+                    full_results = []
+                else:
+                    full_results = []
+                    # TODO - Need to make this iterate through all items in the fixes dict (currently only does the first)
+                    for key, value in fixes.items():    
+                        txt = re.sub('^[0-9]+: ', '', self.text)
+                        loc = txt.find(key) - 1
+                        if self.return_context == "T":
+                            full_results.append('{0} Suggest \'{1}\' for \'{2}\' | {3}'.format(loc, value, key, self.text))
+                        else:
+                            full_results.append('{0} Suggest \'{1}\' for \'{2}\''.format(loc, value, key))
             else:    
                 full_results = [correction, fixes]
             return(full_results)
@@ -335,15 +414,18 @@ class spellcheck:
     def fix(self):
             
         open_list = []
+        
+        # run spellcheck against each paragraph separately
         for i in self._SPLIT_PARAGRAPHS(self.text):
-                open_list.append(spellcheck(i,changes_by_paragraph= self.changes_by_paragraph, interactive = self.interactive, common_scannos = self.common_scannos, top_k = self.top_k).SINGLE_STRING_FIX())          
+            open_list.append(spellcheck(i,changes_by_paragraph= self.changes_by_paragraph, interactive = self.interactive, common_scannos = self.common_scannos, top_k = self.top_k, return_context = self.return_context).SINGLE_STRING_FIX())          
+
         
         if self.changes_by_paragraph == "T":
             open_list = list(filter(None, open_list))
             if len(open_list) == 0:
                 return("NOTE: No changes made to text")
             else:
-                return(open_list)
+                return('\n'.join(sum(open_list, [])))
         else:
             # collapse all corrected paragraphs
             corrections = [x[0] for x in open_list]
@@ -365,15 +447,23 @@ class spellcheck:
                 return(final_text)
 
 
-# TODO - (IGNORE_ALL) add branch to interactive menu which would follow up any suggestions ignored and ask user if they want to Ignore All (for any misspell occurring >2 times in the text)
-# TODO - (UPDATE_ALL) likewise, add branch to interactive menu which would follow up any accepted suggestions that are in the common_scannos list (for any misspell occurring >2 times in the text)
-# TODO - (CLI_DIALOGUE_BOX) make tkinter play nicely with CLI python
-# TODO - (WARM_UP) can we somehow negate the warm-up time for the transformers unmasker? (+ associate warning)?
-# TODO - (IGNORE_SPLIT_WORDS) need to ignore the first word of a new page, since these can be split words across pages (this may also just be tied up in the unsplit functionality, where this word should have a leading * to denote a split word)
+
+
+# TODO - (FULL_PARAGRAPHS) Need to allow BERT context to draw from all lines in a full paragraph (currently resets at each newline -- this corresponds to 1 line of text in a Gutenberg text, and likely leads to degraded spellcheck performance due to loss of context)
+# TODO - (ADD_DICTS) Need to add selectable foreign language dictionaries 
+# TODO - (ADD_STEALTHOS) Need to add additional common stealth scannos to OCRfixr
 # TODO - (FIX_MASHED_WORDS) check for mashed up words ("anhour" --> "an hour") BEFORE concluding they are misspells -- BERT/Spellcheck really can't handle these well, as I quickly found a case where OCRfixr incorrectly changed the text   --->   Walker of the Secret Service book is a great test for this!
     # should be able to leverage symspells compound lookup for this!
+# TODO - (GutenBERT) fine-tune BERT model on Gutenberg texts, to improve relatedness of context suggestions
+# TODO - (WARM_UP) can we somehow negate the warm-up time for the transformers unmasker?
+    # pipelines = 7 secs
+    # symspellpy dictionary load = 3 seconds
+# TODO - (ADD_IGNORES) Need to add additional recurring bad suggestions to Ignore list, based on running OCRfixr on a wide sample of books
+# TODO - (IGNORE_SPLIT_WORDS) need to ignore the first word of a new page, since these can be split words across pages (this may also just be tied up in the unsplit functionality, where this word should have a leading * to denote a split word)
+
 
 # Note: find-replace is not instance-specific, it is paragraph specific..."yov" will be replaced with "you" in all instances found in that section of text. It would be rare, but this may cause issues when a repeated scanno is valid & not valid within the same paragraph
 # Note: OCRfixr ignores all words with leading uppercasing, as these are assumed to be proper nouns, which fall outside of the scope of what this approach can accomplish.
 
+
        
