@@ -29,6 +29,8 @@ class TestStringMethods(unittest.TestCase):
         self.assertEqual(spellcheck("been, that it violates certain axioms above stated, (18,) which have been")._LIST_MISREADS(), [])
         self.assertEqual(spellcheck("been, that it violates certain axioms above stated, (ee) which have been")._LIST_MISREADS(), [])
         self.assertEqual(spellcheck('333.   “The Journal du Magnetisme of the 10th of March, 1853, had')._LIST_MISREADS(), [])
+        self.assertEqual(spellcheck('Be kept open, soft, and tender.” She talked')._LIST_MISREADS(), [])
+        self.assertEqual(spellcheck('of the missionaries to do this.1')._LIST_MISREADS(), [])
 
 
     def test_ignore_words(self):
@@ -41,7 +43,7 @@ class TestStringMethods(unittest.TestCase):
 
 
     def test_finds_easy_errors(self):
-        self.assertEqual(spellcheck("cut the shiit").fix(), "cut the shit")
+        self.assertEqual(spellcheck("cut the sh1t").fix(), "cut the shit")
         self.assertEqual(spellcheck("The birds flevv south").fix(), "The birds flew south")
 
 
@@ -72,25 +74,40 @@ class TestStringMethods(unittest.TestCase):
 
     def test_return_fixes_flag(self):
         self.assertEqual(spellcheck("The birds flevv down\n south",return_fixes = "T").fix(), ["The birds flew down\n south",{("flevv","flew"):1}])
-        self.assertEqual(spellcheck("The birds flevv down\n south and were quikly apprehended",return_fixes = "T").fix(), ["The birds flew down\n south and were quickly apprehended",{("flevv","flew"):1, ("quikly","quickly"):1}])
+        self.assertEqual(spellcheck("The birds flevv down\n south and wefe quickly apprehended",return_fixes = "T").fix(), ["The birds flew down\n south and were quickly apprehended",{("flevv","flew"):1, ("wefe","were"):1}])
 
 
     def test_changes_by_paragraph_flag(self):
-        self.assertEqual(spellcheck("The birds flevv down\n south, bvt wefe quickly apprehended\n by border patrol agents", changes_by_paragraph = "T").fix(), [["The birds flew down\n",{"flevv":"flew"}], [" south, but were quickly apprehended\n", {"bvt":"but", "wefe":"were"}]])
+        self.assertEqual(spellcheck("The birds flevv down\n south, bvt wefe quickly apprehended\n by border patrol agents", changes_by_paragraph = "T").fix(), "9 Suggest 'flew' for 'flevv'\n7 Suggest 'but' for 'bvt'")
         # Case - no misspells in the text
         self.assertEqual(spellcheck("by border patrol agents", changes_by_paragraph = "T").fix(), "NOTE: No changes made to text")
         # Case - misspell in the text, but no replacement
         self.assertEqual(spellcheck("In fact, the effect of circine on the human body\n'", changes_by_paragraph = "T").fix(), "NOTE: No changes made to text")
-    
-        
-    def test_spellcheck_contains_uppercase(self):
-        self.assertEqual(spellcheck("He falls ilL.").fix(), "He falls ill.")
+        # Case - multiple fixes in a single line
+        self.assertEqual(spellcheck("I hope yov will f1nd all the rnistakes in this sentence. Otherwise, I wlll be very sad.", changes_by_paragraph = "T").fix(), "6 Suggest 'you' for 'yov'\n15 Suggest 'find' for 'f1nd'\n69 Suggest 'will' for 'wlll'")
 
 
     def test_fixes_multiple_errors(self):
         self.assertEqual(spellcheck("I hope yov will f1nd all the rnistakes in this sentence. Otherwise, I wlll be very sad.").fix(), "I hope you will find all the rnistakes in this sentence. Otherwise, I will be very sad.")
 
 
+    def test_disregards_homophones(self):
+        # does not change homophone just --> jist
+        self.assertEqual(spellcheck("yuh? You’ll be all right. You’re jist like I was when I begun", changes_by_paragraph = "T").fix(), 'NOTE: No changes made to text')
+        # changes o --> e suggestions (an exception to the homophone rule)
+        self.assertEqual(spellcheck("It is suggested that these words may bo misapprehended. I use them in the sense", changes_by_paragraph = "T").fix(), "36 Suggest 'be' for 'bo'")
+        # does not change suggestions that match manually defined "bad" suggestions 
+        self.assertEqual(spellcheck("and over dere you will see the house", changes_by_paragraph = "T").fix(), 'NOTE: No changes made to text')
+
+
+    def test_fixes_mashed_words(self):
+        self.assertEqual(spellcheck("It seemed a long time as we sat there in the darkness waiting for the train; but it was perhaps, in fact, less than half anhour.", changes_by_paragraph = "T").fix(), "120 Suggest 'an hour' for 'anhour'")
+        # Catches 1 of 2
+        self.assertEqual(spellcheck("happened in such manner that the rumourof itspread throughout", changes_by_paragraph = "T").fix(), "41 Suggest 'it spread' for 'itspread'")
+        # Doesnt separate this one - that's ok
+        self.assertEqual(spellcheck("here, when it was about two hours of the night, a.little more", changes_by_paragraph = "T").fix(), 'NOTE: No changes made to text')
+
+
     def test_spellcheck_speed_acceptable(self):
         # GOALS
         # 0 misspells = < 0.01 seconds  [V1.3 = 0.0025s]
@@ -100,17 +117,6 @@ class TestStringMethods(unittest.TestCase):
         self.assertLessEqual(time_func(sc,"He falls ilL."), 0.2) 
         self.assertLessEqual(time_func(sc,"I hope yov will f1nd all the rnistakes in this sentence. Otherwise, I wlll be very sad."), 1)
 
-# TODO
-
-# MASHUPS
-# This one currently fails - OCRfixr sometimes mis-handles mashed up words
-#    def test_fixes_mashed_words(self):
-#       self.assertEqual(spellcheck("It seemed a long time as we sat there in the darkness waiting for the train; but it was perhaps, in fact, less than half anhour.").fix(), "It seemed a long time as we sat there in the darkness waiting for the train; but it was perhaps, in fact, less than half an hour.")
-
-# EXTRA SPACES
-# This one currently fails - OCRfixr doesn't see the relevant character due to the extra space, and duplicates it
-#    def test_fixes_extra_spaces(self):
-#       self.assertEqual(spellcheck("But suppose I spare your lif e--will you help me to escape?").fix(), "But suppose I spare your life--will you help me to escape?")
 
 
 if __name__ == '__main__':
